"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const url_1 = require("url");
const etag = require("etag");
const onFinished = require("on-finished");
const destroy = require("destroy");
const fs_1 = require("fs");
const path_1 = require("path");
const mime = require("mime-types");
const contentDisposition = require("content-disposition");
const vary = require("vary");
const fresh = require("fresh");
const macroable_1 = require("macroable");
function statFn(filePath) {
    return new Promise((resolve, reject) => {
        fs_1.stat(filePath, (error, stats) => {
            if (error) {
                reject(error);
                return;
            }
            resolve(stats);
        });
    });
}
class Response extends macroable_1.Macroable {
    constructor(request, response, config) {
        super();
        this.request = request;
        this.response = response;
        this.explicitEnd = false;
        this.lazyBody = null;
        this._config = Object.assign({
            etag: false,
            jsonpCallbackName: 'callback',
        }, config);
    }
    get hasLazyBody() {
        return !!(this.lazyBody && this.lazyBody.writer);
    }
    get finished() {
        return this.response.finished;
    }
    get headersSent() {
        return this.response.headersSent;
    }
    get isPending() {
        return (!this.headersSent && !this.finished);
    }
    _castHeaderValue(value) {
        return Array.isArray(value) ? value.map(String) : String(value);
    }
    _writeBody(content, generateEtag, jsonpCallbackName) {
        let { type, body, originalType } = this.buildResponseBody(content);
        if (body === null) {
            this.status(204);
            this.removeHeader('Content-Type');
            this.removeHeader('Content-Length');
            this.removeHeader('Transfer-Encoding');
            this.response.end();
            return;
        }
        if (type === 'unknown') {
            throw new Error(`Cannot send ${originalType} as HTTP response`);
        }
        if ([204, 304].indexOf(this.response.statusCode) > -1) {
            this.removeHeader('Content-Type');
            this.removeHeader('Content-Length');
            this.removeHeader('Transfer-Encoding');
            this.response.end(body);
            return;
        }
        if (jsonpCallbackName) {
            this.header('X-Content-Type-Options', 'nosniff');
            this.safeHeader('Content-Type', 'text/javascript; charset=utf-8');
        }
        else {
            this.safeHeader('Content-Type', `${type}; charset=utf-8`);
        }
        if (generateEtag) {
            this.setEtag(body);
        }
        if (jsonpCallbackName) {
            body = body.replace(/\u2028/g, '\\u2028').replace(/\u2029/g, '\\u2029');
            body = `/**/ typeof ${jsonpCallbackName} === 'function' && ${jsonpCallbackName}(${body});`;
        }
        this.header('Content-Length', Buffer.byteLength(body));
        this.response.write(body);
        this.response.end();
    }
    getHeader(key) {
        return this.response.getHeader(key);
    }
    header(key, value) {
        if (value) {
            this.response.setHeader(key, this._castHeaderValue(value));
        }
        return this;
    }
    append(key, value) {
        if (!value) {
            return this;
        }
        let existingHeader = this.getHeader(key);
        let casted = this._castHeaderValue(value);
        if (!existingHeader) {
            this.response.setHeader(key, casted);
            return this;
        }
        existingHeader = this._castHeaderValue(existingHeader);
        casted = Array.isArray(existingHeader) ? existingHeader.concat(casted) : [existingHeader].concat(casted);
        this.response.setHeader(key, casted);
        return this;
    }
    safeHeader(key, value) {
        if (!this.getHeader(key)) {
            this.header(key, value);
        }
        return this;
    }
    removeHeader(key) {
        this.response.removeHeader(key);
        return this;
    }
    status(code) {
        this.response.statusCode = code;
        return this;
    }
    type(type, charset) {
        type = charset ? `${type}; charset=${charset}` : type;
        this.header('Content-Type', mime.contentType(type));
        return this;
    }
    vary(field) {
        vary(this.response, field);
        return this;
    }
    setEtag(body, weak = false) {
        this.header('Etag', etag(body, { weak }));
        return this;
    }
    fresh() {
        if (this.request.method && ['GET', 'HEAD'].indexOf(this.request.method) === -1) {
            return false;
        }
        const status = this.response.statusCode;
        if ((status >= 200 && status < 300) || status === 304) {
            return fresh(this.request.headers, this.response.getHeaders());
        }
        return false;
    }
    buildResponseBody(body) {
        if (!body) {
            return {
                body: null,
                type: 'null',
            };
        }
        if (typeof (body) === 'string') {
            return {
                body,
                type: /^\s*</.test(body) ? 'text/html' : 'text/plain',
            };
        }
        if (Buffer.isBuffer(body)) {
            return {
                body,
                type: 'application/octet-stream',
            };
        }
        if (typeof (body) === 'number' || typeof (body) === 'boolean') {
            return {
                body: String(body),
                type: 'text/plain',
            };
        }
        if (typeof (body) === 'object') {
            return {
                body: JSON.stringify(body),
                type: 'application/json',
            };
        }
        return {
            body,
            originalType: typeof (body),
            type: 'unknown',
        };
    }
    send(body, generateEtag = this._config.etag) {
        if (this.explicitEnd) {
            this.lazyBody = {
                writer: this._writeBody,
                args: [body, generateEtag],
            };
            return;
        }
        this._writeBody(body, generateEtag);
    }
    json(body, generateEtag) {
        return this.send(body, generateEtag);
    }
    jsonp(body, callbackName = this._config.jsonpCallbackName, generateEtag = this._config.etag) {
        if (this.explicitEnd) {
            this.lazyBody = {
                writer: this._writeBody,
                args: [body, generateEtag, callbackName],
            };
            return;
        }
        this._writeBody(body, generateEtag, callbackName);
    }
    stream(body, raiseErrors = false) {
        if (typeof (body.pipe) !== 'function' || !body.readable || typeof (body.read) !== 'function') {
            throw new Error('response.stream accepts a readable stream only');
        }
        return new Promise((resolve, reject) => {
            this.explicitEnd = false;
            let finished = false;
            body.on('error', (error) => {
                if (finished) {
                    return;
                }
                finished = true;
                destroy(body);
                if (raiseErrors) {
                    reject(error);
                }
                else {
                    this.status(error.status || 500);
                    this.response.end(error.code === 'ENOENT' ? 'File not found' : 'Cannot process file');
                    resolve();
                }
            });
            body.on('end', resolve);
            onFinished(this.response, () => {
                finished = true;
                destroy(body);
            });
            body.pipe(this.response);
        });
    }
    async download(filePath, generateEtag = this._config.etag, raiseErrors = false) {
        this.explicitEnd = false;
        try {
            const stats = await statFn(filePath);
            if (!stats || !stats.isFile()) {
                throw new Error('response.download only accepts path to a file');
            }
            this.header('Last-Modified', stats.mtime.toUTCString());
            this.header('Content-length', stats.size);
            this.type(path_1.extname(filePath));
            if (!generateEtag) {
                if (this.request.method === 'HEAD') {
                    this.status(200);
                    this.response.end();
                    return;
                }
                return this.stream(fs_1.createReadStream(filePath), raiseErrors);
            }
            this.setEtag(stats, true);
            const fresh = this.fresh();
            if (fresh) {
                this.status(304);
                this.response.end();
                return;
            }
            if (this.request.method === 'HEAD') {
                this.status(200);
                this.response.end();
                return;
            }
            return this.stream(fs_1.createReadStream(filePath), raiseErrors);
        }
        catch (error) {
            if (raiseErrors) {
                throw error;
            }
            else {
                this.status(404);
                this.response.end('Cannot process file');
            }
        }
    }
    async attachment(filePath, name, disposition, generateEtag, raiseErrors) {
        name = name || filePath;
        this.header('Content-Disposition', contentDisposition(name, { type: disposition }));
        return this.download(filePath, generateEtag, raiseErrors);
    }
    location(url) {
        this.header('Location', url);
        return this;
    }
    redirect(url, sendQueryParams, statusCode = 302) {
        url = sendQueryParams ? `${url}?${url_1.parse(this.request.url, false).query}` : url;
        this.explicitEnd = false;
        this.location(url);
        this.status(statusCode);
        this.response.end(`Redirecting to ${url}`);
    }
    finish() {
        if (this.explicitEnd && this.lazyBody && this.isPending) {
            this.lazyBody.writer.bind(this)(...this.lazyBody.args);
            this.lazyBody = null;
        }
    }
}
Response._macros = {};
Response._getters = {};
exports.Response = Response;
